# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Collection configuration file.
"""

import configparser

FILENAME = '.xandikos'


class CollectionMetadata(object):
    """Metadata for a configuration."""

    def get_color(self) -> str:
        """Get the color for this collection.
        """
        raise NotImplementedError(self.get_color)

    def set_color(self, color: str) -> None:
        """Change the color of this collection."""
        raise NotImplementedError(self.set_color)

    def get_source_url(self) -> str:
        """Get the source URL for this collection.
        """
        raise NotImplementedError(self.get_source_url)

    def set_source_url(self, url: str) -> None:
        """Set the source URL for this collection.
        """
        raise NotImplementedError(self.set_source_url)

    def get_comment(self) -> str:
        raise NotImplementedError(self.get_comment)

    def get_displayname(self) -> str:
        raise NotImplementedError(self.get_displayname)

    def get_description(self) -> str:
        raise NotImplementedError(self.get_description)

    def get_order(self) -> str:
        raise NotImplementedError(self.get_order)

    def set_order(self, order: str) -> None:
        raise NotImplementedError(self.set_order)


class FileBasedCollectionMetadata(CollectionMetadata):
    """Metadata for a configuration."""

    def __init__(self, cp=None, save=None):
        if cp is None:
            cp = configparser.ConfigParser()
        self._configparser = cp
        self._save_cb = save

    def _save(self, message):
        if self._save_cb is None:
            return
        self._save_cb(self._configparser, message)

    @classmethod
    def from_file(cls, f):
        cp = configparser.ConfigParser()
        cp.read_file(f)
        return cls(cp)

    def get_source_url(self):
        return self._configparser['DEFAULT']['source']

    def set_source_url(self, url):
        if url is not None:
            self._configparser['DEFAULT']['source'] = url
        else:
            del self._configparser['DEFAULT']['source']
        self._save("Set source URL.")

    def get_color(self):
        return self._configparser['DEFAULT']['color']

    def get_comment(self):
        return self._configparser['DEFAULT']['comment']

    def get_displayname(self):
        return self._configparser['DEFAULT']['displayname']

    def get_description(self):
        return self._configparser['DEFAULT']['description']

    def set_color(self, color):
        if color is not None:
            self._configparser['DEFAULT']['color'] = color
        else:
            del self._configparser['DEFAULT']['color']
        self._save("Set color.")

    def set_displayname(self, displayname):
        if displayname is not None:
            self._configparser['DEFAULT']['displayname'] = displayname
        else:
            del self._configparser['DEFAULT']['displayname']
        self._save("Set display name.")

    def set_description(self, description):
        if description is not None:
            self._configparser['DEFAULT']['description'] = description
        else:
            del self._configparser['DEFAULT']['description']
        self._save("Set description.")

    def set_comment(self, comment):
        if comment is not None:
            self._configparser['DEFAULT']['comment'] = comment
        else:
            del self._configparser['DEFAULT']['comment']
        self._save("Set comment.")

    def set_type(self, store_type):
        self._configparser['DEFAULT']['type'] = store_type
        self._save("Set collection type.")

    def get_type(self):
        return self._configparser['DEFAULT']['type']

    def get_order(self):
        return self._configparser['calendar']['order']

    def set_order(self, order):
        try:
            self._configparser.add_section('calendar')
        except configparser.DuplicateSectionError:
            pass
        if order is None:
            del self._configparser['calendar']['order']
        else:
            self._configparser['calendar']['order'] = order



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Tests for xandikos.icalendar."""

from datetime import datetime

import pytz
import unittest

from icalendar.cal import Event

from xandikos import (
    collation as _mod_collation,
)
from xandikos.icalendar import (
    CalendarFilter,
    ICalendarFile,
    MissingProperty,
    TextMatcher,
    validate_calendar,
    apply_time_range_vevent,
    as_tz_aware_ts,
)
from xandikos.store import InvalidFileContents

EXAMPLE_VCALENDAR1 = b"""\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//bitfire web engineering//DAVdroid 0.8.0 (ical4j 1.0.x)//EN
BEGIN:VTODO
CREATED:20150314T223512Z
DTSTAMP:20150527T221952Z
LAST-MODIFIED:20150314T223512Z
STATUS:NEEDS-ACTION
SUMMARY:do something
UID:bdc22720-b9e1-42c9-89c2-a85405d8fbff
END:VTODO
END:VCALENDAR
"""

EXAMPLE_VCALENDAR_WITH_PARAM = b"""\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//bitfire web engineering//DAVdroid 0.8.0 (ical4j 1.0.x)//EN
BEGIN:VTODO
CREATED;TZID=America/Denver:20150314T223512Z
DTSTAMP:20150527T221952Z
LAST-MODIFIED:20150314T223512Z
STATUS:NEEDS-ACTION
SUMMARY:do something
UID:bdc22720-b9e1-42c9-89c2-a85405d8fbff
END:VTODO
END:VCALENDAR
"""

EXAMPLE_VCALENDAR_NO_UID = b"""\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//bitfire web engineering//DAVdroid 0.8.0 (ical4j 1.0.x)//EN
BEGIN:VTODO
CREATED:20120314T223512Z
DTSTAMP:20130527T221952Z
LAST-MODIFIED:20150314T223512Z
STATUS:NEEDS-ACTION
SUMMARY:do something without uid
END:VTODO
END:VCALENDAR
"""

EXAMPLE_VCALENDAR_INVALID_CHAR = b"""\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//bitfire web engineering//DAVdroid 0.8.0 (ical4j 1.0.x)//EN
BEGIN:VTODO
CREATED:20150314T223512Z
DTSTAMP:20150527T221952Z
LAST-MODIFIED:20150314T223512Z
STATUS:NEEDS-ACTION
SUMMARY:do something
ID:bdc22720-b9e1-42c9-89c2-a85405d8fbff
END:VTODO
END:VCALENDAR
"""


class ExtractCalendarUIDTests(unittest.TestCase):

    def test_extract_str(self):
        fi = ICalendarFile([EXAMPLE_VCALENDAR1], 'text/calendar')
        self.assertEqual(
            'bdc22720-b9e1-42c9-89c2-a85405d8fbff',
            fi.get_uid())
        fi.validate()

    def test_extract_no_uid(self):
        fi = ICalendarFile([EXAMPLE_VCALENDAR_NO_UID], 'text/calendar')
        fi.validate()
        self.assertEqual(["Missing required field UID"],
                         list(validate_calendar(fi.calendar, strict=True)))
        self.assertEqual([],
                         list(validate_calendar(fi.calendar, strict=False)))
        self.assertRaises(KeyError, fi.get_uid)

    def test_invalid_character(self):
        fi = ICalendarFile([EXAMPLE_VCALENDAR_INVALID_CHAR], 'text/calendar')
        self.assertRaises(InvalidFileContents, fi.validate)
        self.assertEqual(["Invalid character b'\\\\x0c' in field SUMMARY"],
                         list(validate_calendar(fi.calendar, strict=False)))


class CalendarFilterTests(unittest.TestCase):

    def setUp(self):
        self.cal = ICalendarFile([EXAMPLE_VCALENDAR1], 'text/calendar')

    def test_simple_comp_filter(self):
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent('VEVENT')
        self.assertEqual(filter.index_keys(), [['C=VCALENDAR/C=VEVENT']])
        self.assertEqual(
            self.cal.get_indexes(
                ['C=VCALENDAR/C=VEVENT', 'C=VCALENDAR/C=VTODO']),
            {'C=VCALENDAR/C=VEVENT': [], 'C=VCALENDAR/C=VTODO': [True]})
        self.assertFalse(
            filter.check_from_indexes(
                'file', {'C=VCALENDAR/C=VEVENT': [],
                         'C=VCALENDAR/C=VTODO': [True]}))
        self.assertFalse(filter.check('file', self.cal))
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent('VTODO')
        self.assertTrue(filter.check('file', self.cal))
        self.assertTrue(
            filter.check_from_indexes(
                'file', {'C=VCALENDAR/C=VEVENT': [],
                         'C=VCALENDAR/C=VTODO': [True]}))

    def test_simple_comp_missing_filter(self):
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO', is_not_defined=True)
        self.assertEqual(
            filter.index_keys(), [['C=VCALENDAR/C=VTODO'], ['C=VCALENDAR']])
        self.assertFalse(
            filter.check_from_indexes(
                'file', {
                    'C=VCALENDAR': [True],
                    'C=VCALENDAR/C=VEVENT': [],
                    'C=VCALENDAR/C=VTODO': [True]}))
        self.assertFalse(filter.check('file', self.cal))
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VEVENT', is_not_defined=True)
        self.assertTrue(filter.check('file', self.cal))
        self.assertTrue(
            filter.check_from_indexes(
                'file', {
                    'C=VCALENDAR': [True],
                    'C=VCALENDAR/C=VEVENT': [],
                    'C=VCALENDAR/C=VTODO': [True]}))

    def test_prop_presence_filter(self):
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_property('X-SUMMARY')
        self.assertEqual(
            filter.index_keys(),
            [['C=VCALENDAR/C=VTODO/P=X-SUMMARY']])
        self.assertFalse(
            filter.check_from_indexes(
                'file', {'C=VCALENDAR/C=VTODO/P=X-SUMMARY': []}))
        self.assertFalse(filter.check('file', self.cal))
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_property('SUMMARY')
        self.assertTrue(
            filter.check_from_indexes(
                'file', {'C=VCALENDAR/C=VTODO/P=SUMMARY': [b'do something']}))
        self.assertTrue(filter.check('file', self.cal))

    def test_prop_explicitly_missing_filter(self):
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VEVENT').filter_property('X-SUMMARY', is_not_defined=True)
        self.assertEqual(
            filter.index_keys(),
            [['C=VCALENDAR/C=VEVENT/P=X-SUMMARY'], ['C=VCALENDAR/C=VEVENT']])
        self.assertFalse(
            filter.check_from_indexes(
                'file',
                {'C=VCALENDAR/C=VEVENT/P=X-SUMMARY': [],
                 'C=VCALENDAR/C=VEVENT': []}))
        self.assertFalse(filter.check('file', self.cal))
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_property('X-SUMMARY', is_not_defined=True)
        self.assertTrue(
            filter.check_from_indexes(
                'file', {
                    'C=VCALENDAR/C=VTODO/P=X-SUMMARY': [],
                    'C=VCALENDAR/C=VTODO': [True]}))
        self.assertTrue(filter.check('file', self.cal))

    def test_prop_text_match(self):
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_property('SUMMARY').filter_text_match(
                b'do something different')
        self.assertEqual(
            filter.index_keys(),
            [['C=VCALENDAR/C=VTODO/P=SUMMARY']])
        self.assertFalse(
            filter.check_from_indexes(
                'file', {'C=VCALENDAR/C=VTODO/P=SUMMARY': [b'do something']}))
        self.assertFalse(filter.check('file', self.cal))
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_property('SUMMARY').filter_text_match(
                b'do something')
        self.assertTrue(
            filter.check_from_indexes(
                'file', {'C=VCALENDAR/C=VTODO/P=SUMMARY': [b'do something']}))
        self.assertTrue(filter.check('file', self.cal))

    def test_param_text_match(self):
        self.cal = ICalendarFile(
            [EXAMPLE_VCALENDAR_WITH_PARAM], 'text/calendar')
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_property('CREATED').filter_parameter(
                'TZID').filter_text_match(
                b'America/Blah')
        self.assertEqual(
            filter.index_keys(),
            [['C=VCALENDAR/C=VTODO/P=CREATED/A=TZID'],
             ['C=VCALENDAR/C=VTODO/P=CREATED']])
        self.assertFalse(
            filter.check_from_indexes(
                'file',
                {'C=VCALENDAR/C=VTODO/P=CREATED/A=TZID': [b'America/Denver']}))
        self.assertFalse(filter.check('file', self.cal))
        filter = CalendarFilter(None)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_property('CREATED').filter_parameter(
                'TZID').filter_text_match(
                b'America/Denver')
        self.assertTrue(
            filter.check_from_indexes(
                'file',
                {'C=VCALENDAR/C=VTODO/P=CREATED/A=TZID': [b'America/Denver']}))
        self.assertTrue(filter.check('file', self.cal))

    def _tzify(self, dt):
        return as_tz_aware_ts(dt, pytz.utc)

    def test_prop_apply_time_range(self):
        filter = CalendarFilter(self._tzify)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_property('CREATED').filter_time_range(
                self._tzify(datetime(2019, 3, 10, 22, 35, 12)),
                self._tzify(datetime(2019, 3, 18, 22, 35, 12)))
        self.assertEqual(
            filter.index_keys(),
            [['C=VCALENDAR/C=VTODO/P=CREATED']])
        self.assertFalse(
            filter.check_from_indexes(
                'file',
                {'C=VCALENDAR/C=VTODO/P=CREATED': ['20150314T223512Z']}))
        self.assertFalse(filter.check('file', self.cal))
        filter = CalendarFilter(self._tzify)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_property('CREATED').filter_time_range(
                self._tzify(datetime(2015, 3, 10, 22, 35, 12)),
                self._tzify(datetime(2015, 3, 18, 22, 35, 12)))
        self.assertTrue(
            filter.check_from_indexes(
                'file',
                {'C=VCALENDAR/C=VTODO/P=CREATED': ['20150314T223512Z']}))
        self.assertTrue(filter.check('file', self.cal))

    def test_comp_apply_time_range(self):
        filter = CalendarFilter(self._tzify)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_time_range(
                self._tzify(datetime(2015, 3, 3, 22, 35, 12)),
                self._tzify(datetime(2015, 3, 10, 22, 35, 12)))
        self.assertEqual(
            filter.index_keys(),
            [['C=VCALENDAR/C=VTODO/P=DTSTART'],
             ['C=VCALENDAR/C=VTODO/P=DUE'],
             ['C=VCALENDAR/C=VTODO/P=DURATION'],
             ['C=VCALENDAR/C=VTODO/P=CREATED'],
             ['C=VCALENDAR/C=VTODO/P=COMPLETED'],
             ['C=VCALENDAR/C=VTODO']])
        self.assertFalse(
            filter.check_from_indexes(
                'file',
                {'C=VCALENDAR/C=VTODO/P=CREATED': ['20150314T223512Z'],
                 'C=VCALENDAR/C=VTODO': [True],
                 'C=VCALENDAR/C=VTODO/P=DUE': [],
                 'C=VCALENDAR/C=VTODO/P=DURATION': [],
                 'C=VCALENDAR/C=VTODO/P=COMPLETED': [],
                 'C=VCALENDAR/C=VTODO/P=DTSTART': [],
                 }))
        self.assertFalse(filter.check('file', self.cal))
        filter = CalendarFilter(self._tzify)
        filter.filter_subcomponent('VCALENDAR').filter_subcomponent(
            'VTODO').filter_time_range(
                self._tzify(datetime(2015, 3, 10, 22, 35, 12)),
                self._tzify(datetime(2015, 3, 18, 22, 35, 12)))
        self.assertTrue(
            filter.check_from_indexes(
                'file',
                {'C=VCALENDAR/C=VTODO/P=CREATED': ['20150314T223512Z'],
                 'C=VCALENDAR/C=VTODO': [True],
                 'C=VCALENDAR/C=VTODO/P=DUE': [],
                 'C=VCALENDAR/C=VTODO/P=DURATION': [],
                 'C=VCALENDAR/C=VTODO/P=COMPLETED': [],
                 'C=VCALENDAR/C=VTODO/P=DTSTART': [],
                 }))
        self.assertTrue(filter.check('file', self.cal))


class TextMatchTest(unittest.TestCase):

    def test_default_collation(self):
        tm = TextMatcher(b"foobar")
        self.assertTrue(tm.match(b"FOOBAR"))
        self.assertTrue(tm.match(b"foobar"))
        self.assertFalse(tm.match(b"fobar"))

    def test_casecmp_collation(self):
        tm = TextMatcher(b'foobar', collation='i;ascii-casemap')
        self.assertTrue(tm.match(b"FOOBAR"))
        self.assertTrue(tm.match(b"foobar"))
        self.assertFalse(tm.match(b"fobar"))

    def test_cmp_collation(self):
        tm = TextMatcher(b'foobar', 'i;octet')
        self.assertFalse(tm.match(b"FOOBAR"))
        self.assertTrue(tm.match(b"foobar"))
        self.assertFalse(tm.match(b"fobar"))

    def test_unknown_collation(self):
        self.assertRaises(
            _mod_collation.UnknownCollation, TextMatcher,
            b'foobar', collation='i;blah')


class ApplyTimeRangeVeventTests(unittest.TestCase):

    def _tzify(self, dt):
        return as_tz_aware_ts(dt, 'UTC')

    def test_missing_dtstart(self):
        ev = Event()
        self.assertRaises(
            MissingProperty, apply_time_range_vevent,
            datetime.utcnow(), datetime.utcnow(), ev, self._tzify)



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

from icalendar.cal import Calendar as ICalendar
import unittest
from wsgiref.util import setup_testing_defaults

from xandikos import caldav
from xandikos.webdav import Property, WebDAVApp, ET

from xandikos.tests import test_webdav


class WebTests(test_webdav.WebTestCase):

    def makeApp(self, backend):
        app = WebDAVApp(backend)
        app.register_methods([caldav.MkcalendarMethod()])
        return app

    def mkcalendar(self, app, path):
        environ = {'PATH_INFO': path, 'REQUEST_METHOD': 'MKCALENDAR',
                   'SCRIPT_NAME': ''}
        setup_testing_defaults(environ)
        _code = []
        _headers = []

        def start_response(code, headers):
            _code.append(code)
            _headers.extend(headers)
        contents = b''.join(app(environ, start_response))
        return _code[0], _headers, contents

    def test_mkcalendar_ok(self):
        class Backend(object):
            def create_collection(self, relpath):
                pass

            def get_resource(self, relpath):
                return None

        class ResourceTypeProperty(Property):
            name = '{DAV:}resourcetype'

            async def get_value(unused_self, href, resource, ret, environ):
                ET.SubElement(ret, '{DAV:}collection')

            def set_value(unused_self, href, resource, ret):
                self.assertEqual(
                    ['{DAV:}collection',
                     '{urn:ietf:params:xml:ns:caldav}calendar'],
                    [x.tag for x in ret])

        app = self.makeApp(Backend())
        app.register_properties([ResourceTypeProperty()])
        code, headers, contents = self.mkcalendar(app, '/resource/bla')
        self.assertEqual('201 Created', code)
        self.assertEqual(b'', contents)


class ExtractfromCalendarTests(unittest.TestCase):

    def setUp(self):
        super(ExtractfromCalendarTests, self).setUp()
        self.requested = ET.Element('{%s}calendar-data' % caldav.NAMESPACE)

    def extractEqual(self, incal_str, outcal_str):
        incal = ICalendar.from_ical(incal_str)
        expected_outcal = ICalendar.from_ical(outcal_str)
        outcal = ICalendar()
        caldav.extract_from_calendar(incal, outcal, self.requested)
        self.assertEqual(
            expected_outcal.to_ical().decode(),
            outcal.to_ical().decode(),
            ET.tostring(self.requested))

    def test_comp(self):
        comp = ET.SubElement(self.requested, '{%s}comp' % caldav.NAMESPACE)
        comp.set('name', 'VCALENDAR')
        self.extractEqual("""\
BEGIN:VCALENDAR
BEGIN:VTODO
CLASS:PUBLIC
COMPLETED:20100829T234417Z
CREATED:20090606T042958Z
END:VTODO
END:VCALENDAR
""", """\
BEGIN:VCALENDAR
END:VCALENDAR
""")

    def test_comp_nested(self):
        vcal_comp = ET.SubElement(
            self.requested, '{%s}comp' % caldav.NAMESPACE)
        vcal_comp.set('name', 'VCALENDAR')
        vtodo_comp = ET.SubElement(vcal_comp, '{%s}comp' % caldav.NAMESPACE)
        vtodo_comp.set('name', 'VTODO')
        self.extractEqual("""\
BEGIN:VCALENDAR
BEGIN:VTODO
COMPLETED:20100829T234417Z
CREATED:20090606T042958Z
END:VTODO
END:VCALENDAR
""", """\
BEGIN:VCALENDAR
BEGIN:VTODO
END:VTODO
END:VCALENDAR
""")
        self.extractEqual("""\
BEGIN:VCALENDAR
BEGIN:VEVENT
COMPLETED:20100829T234417Z
CREATED:20090606T042958Z
END:VEVENT
END:VCALENDAR
""", """\
BEGIN:VCALENDAR
END:VCALENDAR
""")

    def test_prop(self):
        vcal_comp = ET.SubElement(
            self.requested, '{%s}comp' % caldav.NAMESPACE)
        vcal_comp.set('name', 'VCALENDAR')
        vtodo_comp = ET.SubElement(vcal_comp, '{%s}comp' % caldav.NAMESPACE)
        vtodo_comp.set('name', 'VTODO')
        completed_prop = ET.SubElement(
            vtodo_comp, '{%s}prop' % caldav.NAMESPACE)
        completed_prop.set('name', 'COMPLETED')
        self.extractEqual("""\
BEGIN:VCALENDAR
BEGIN:VTODO
COMPLETED:20100829T234417Z
CREATED:20090606T042958Z
END:VTODO
END:VCALENDAR
""", """\
BEGIN:VCALENDAR
BEGIN:VTODO
COMPLETED:20100829T234417Z
END:VTODO
END:VCALENDAR
""")
        self.extractEqual("""\
BEGIN:VCALENDAR
BEGIN:VEVENT
CREATED:20090606T042958Z
END:VEVENT
END:VCALENDAR
""", """\
BEGIN:VCALENDAR
END:VCALENDAR
""")

    def test_allprop(self):
        vcal_comp = ET.SubElement(
            self.requested, '{%s}comp' % caldav.NAMESPACE)
        vcal_comp.set('name', 'VCALENDAR')
        vtodo_comp = ET.SubElement(vcal_comp, '{%s}comp' % caldav.NAMESPACE)
        vtodo_comp.set('name', 'VTODO')
        ET.SubElement(vtodo_comp, '{%s}allprop' % caldav.NAMESPACE)
        self.extractEqual("""\
BEGIN:VCALENDAR
BEGIN:VTODO
COMPLETED:20100829T234417Z
CREATED:20090606T042958Z
END:VTODO
END:VCALENDAR
""", """\
BEGIN:VCALENDAR
BEGIN:VTODO
COMPLETED:20100829T234417Z
CREATED:20090606T042958Z
END:VTODO
END:VCALENDAR
""")

    def test_allcomp(self):
        vcal_comp = ET.SubElement(
            self.requested,
            '{%s}comp' % caldav.NAMESPACE)
        vcal_comp.set('name', 'VCALENDAR')
        ET.SubElement(vcal_comp, '{%s}allcomp' % caldav.NAMESPACE)
        self.extractEqual("""\
BEGIN:VCALENDAR
BEGIN:VTODO
COMPLETED:20100829T234417Z
CREATED:20090606T042958Z
END:VTODO
END:VCALENDAR
""", """\
BEGIN:VCALENDAR
BEGIN:VTODO
END:VTODO
END:VCALENDAR
""")

    def test_expand(self):
        expand = ET.SubElement(self.requested, '{%s}expand' % caldav.NAMESPACE)
        expand.set('start', '20060103T000000Z')
        expand.set('end', '20060105T000000Z')
        with self.assertRaises(NotImplementedError):
            self.extractEqual("""\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//Example Corp.//CalDAV Client//EN
BEGIN:VTIMEZONE
LAST-MODIFIED:20040110T032845Z
TZID:US/Eastern
BEGIN:DAYLIGHT
DTSTART:20000404T020000
RRULE:FREQ=YEARLY;BYDAY=1SU;BYMONTH=4
TZNAME:EDT
TZOFFSETFROM:-0500
TZOFFSETTO:-0400
END:DAYLIGHT
BEGIN:STANDARD
DTSTART:20001026T020000
RRULE:FREQ=YEARLY;BYDAY=-1SU;BYMONTH=10
TZNAME:EST
TZOFFSETFROM:-0400
TZOFFSETTO:-0500
END:STANDARD
END:VTIMEZONE
BEGIN:VEVENT
DTSTAMP:20060206T001121Z
DTSTART;TZID=US/Eastern:20060102T120000
DURATION:PT1H
RRULE:FREQ=DAILY;COUNT=5
SUMMARY:Event #2
UID:00959BC664CA650E933C892C@example.com
END:VEVENT
BEGIN:VEVENT
DTSTAMP:20060206T001121Z
DTSTART;TZID=US/Eastern:20060104T140000
DURATION:PT1H
RECURRENCE-ID;TZID=US/Eastern:20060104T120000
SUMMARY:Event #2 bis
UID:00959BC664CA650E933C892C@example.com
END:VEVENT
END:VCALENDAR
""", """\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//Example Corp.//CalDAV Client//EN
BEGIN:VEVENT
DTSTAMP:20060206T001121Z
DTSTART:20060103T170000
DURATION:PT1H
RECURRENCE-ID:20060103T170000
SUMMARY:Event #2
UID:00959BC664CA650E933C892C@example.com
END:VEVENT
BEGIN:VEVENT
DTSTAMP:20060206T001121Z
DTSTART:20060104T190000
DURATION:PT1H
RECURRENCE-ID:20060104T170000
SUMMARY:Event #2 bis
UID:00959BC664CA650E933C892C@example.com
END:VEVENT
END:VCALENDAR
""")



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Tests for xandikos.web."""

import os
import shutil
import tempfile
import unittest

from .. import caldav
from ..icalendar import ICalendarFile
from ..store.vdir import VdirStore
from ..web import (
    XandikosBackend,
    CalendarCollection,
)


EXAMPLE_VCALENDAR1 = b"""\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//bitfire web engineering//DAVdroid 0.8.0 (ical4j 1.0.x)//EN
BEGIN:VTODO
CREATED:20150314T223512Z
DTSTAMP:20150527T221952Z
LAST-MODIFIED:20150314T223512Z
STATUS:NEEDS-ACTION
SUMMARY:do something
UID:bdc22720-b9e1-42c9-89c2-a85405d8fbff
END:VTODO
END:VCALENDAR
"""


class CalendarCollectionTests(unittest.TestCase):

    def setUp(self):
        super(CalendarCollectionTests, self).setUp()
        self.tempdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tempdir)

        self.store = VdirStore.create(os.path.join(self.tempdir, 'c'))
        self.store.load_extra_file_handler(ICalendarFile)
        self.backend = XandikosBackend(self.tempdir)

        self.cal = CalendarCollection(self.backend, 'c', self.store)

    def test_description(self):
        self.store.set_description('foo')
        self.assertEqual('foo', self.cal.get_calendar_description())

    def test_color(self):
        self.assertRaises(KeyError, self.cal.get_calendar_color)
        self.cal.set_calendar_color('#aabbcc')
        self.assertEqual('#aabbcc', self.cal.get_calendar_color())

    def test_get_supported_calendar_components(self):
        self.assertEqual(
            ["VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY"],
            self.cal.get_supported_calendar_components())

    def test_calendar_query_vtodos(self):
        def create_fn(cls):
            f = cls(None)
            f.filter_subcomponent('VCALENDAR').filter_subcomponent('VTODO')
            return f
        self.assertEqual([], list(self.cal.calendar_query(create_fn)))
        self.store.import_one('foo.ics', 'text/calendar', [EXAMPLE_VCALENDAR1])
        result = list(self.cal.calendar_query(create_fn))
        self.assertEqual(1, len(result))
        self.assertEqual('foo.ics', result[0][0])
        self.assertIs(self.store, result[0][1].store)
        self.assertEqual('foo.ics', result[0][1].name)
        self.assertEqual('text/calendar', result[0][1].content_type)

    def test_calendar_query_vtodo_by_uid(self):
        def create_fn(cls):
            f = cls(None)
            f.filter_subcomponent(
                'VCALENDAR').filter_subcomponent(
                'VTODO').filter_property(
                'UID').filter_text_match(
                    b'bdc22720-b9e1-42c9-89c2-a85405d8fbff')
            return f
        self.assertEqual([], list(self.cal.calendar_query(create_fn)))
        self.store.import_one('foo.ics', 'text/calendar', [EXAMPLE_VCALENDAR1])
        result = list(self.cal.calendar_query(create_fn))
        self.assertEqual(1, len(result))
        self.assertEqual('foo.ics', result[0][0])
        self.assertIs(self.store, result[0][1].store)
        self.assertEqual('foo.ics', result[0][1].name)
        self.assertEqual('text/calendar', result[0][1].content_type)

    def test_get_supported_calendar_data_types(self):
        self.assertEqual(
            [('text/calendar', '1.0'), ('text/calendar', '2.0')],
            self.cal.get_supported_calendar_data_types())

    def test_get_max_date_time(self):
        self.assertEqual(
            "99991231T235959Z", self.cal.get_max_date_time())

    def test_get_min_date_time(self):
        self.assertEqual(
            "00010101T000000Z", self.cal.get_min_date_time())

    def test_members(self):
        self.assertEqual([], list(self.cal.members()))
        self.store.import_one('foo.ics', 'text/calendar', [EXAMPLE_VCALENDAR1])
        result = list(self.cal.members())
        self.assertEqual(1, len(result))
        self.assertEqual('foo.ics', result[0][0])
        self.assertIs(self.store, result[0][1].store)
        self.assertEqual('foo.ics', result[0][1].name)
        self.assertEqual('text/calendar', result[0][1].content_type)

    def test_get_member(self):
        self.assertRaises(KeyError, self.cal.get_member, 'foo.ics')
        self.store.import_one('foo.ics', 'text/calendar', [EXAMPLE_VCALENDAR1])
        result = self.cal.get_member('foo.ics')
        self.assertIs(self.store, result.store)
        self.assertEqual('foo.ics', result.name)
        self.assertEqual('text/calendar', result.content_type)

    def test_delete_member(self):
        self.assertRaises(KeyError, self.cal.get_member, 'foo.ics')
        self.store.import_one('foo.ics', 'text/calendar', [EXAMPLE_VCALENDAR1])
        self.cal.get_member('foo.ics')
        self.cal.delete_member('foo.ics')
        self.assertRaises(KeyError, self.cal.get_member, 'foo.ics')

    def test_get_schedule_calendar_transparency(self):
        self.assertEqual(
            caldav.TRANSPARENCY_OPAQUE,
            self.cal.get_schedule_calendar_transparency())



# Xandikos
# Copyright (C) 2018 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Tests for xandikos.store.config."""

from io import StringIO

from unittest import TestCase

import dulwich.repo
from ..store.config import FileBasedCollectionMetadata
from ..store.git import RepoCollectionMetadata


class FileBasedCollectionMetadataTests(TestCase):

    def test_get_color(self):
        f = StringIO("""\
[DEFAULT]
color = #ffffff
""")
        cc = FileBasedCollectionMetadata.from_file(f)
        self.assertEqual('#ffffff', cc.get_color())

    def test_get_color_missing(self):
        f = StringIO("")
        cc = FileBasedCollectionMetadata.from_file(f)
        self.assertRaises(KeyError, cc.get_color)

    def test_get_comment(self):
        f = StringIO("""\
[DEFAULT]
comment = this is a comment
""")
        cc = FileBasedCollectionMetadata.from_file(f)
        self.assertEqual('this is a comment', cc.get_comment())

    def test_get_comment_missing(self):
        f = StringIO("")
        cc = FileBasedCollectionMetadata.from_file(f)
        self.assertRaises(KeyError, cc.get_comment)

    def test_get_description(self):
        f = StringIO("""\
[DEFAULT]
description = this is a description
""")
        cc = FileBasedCollectionMetadata.from_file(f)
        self.assertEqual('this is a description', cc.get_description())

    def test_get_description_missing(self):
        f = StringIO("")
        cc = FileBasedCollectionMetadata.from_file(f)
        self.assertRaises(KeyError, cc.get_description)

    def test_get_displayname(self):
        f = StringIO("""\
[DEFAULT]
displayname = DISPLAY-NAME
""")
        cc = FileBasedCollectionMetadata.from_file(f)
        self.assertEqual('DISPLAY-NAME', cc.get_displayname())

    def test_get_displayname_missing(self):
        f = StringIO("")
        cc = FileBasedCollectionMetadata.from_file(f)
        self.assertRaises(KeyError, cc.get_displayname)


class MetadataTests(object):

    def test_color(self):
        self.assertRaises(KeyError, self._config.get_color)
        self._config.set_color('#ffffff')
        self.assertEqual('#ffffff', self._config.get_color())
        self._config.set_color(None)
        self.assertRaises(KeyError, self._config.get_color)

    def test_comment(self):
        self.assertRaises(KeyError, self._config.get_comment)
        self._config.set_comment('this is a comment')
        self.assertEqual('this is a comment', self._config.get_comment())
        self._config.set_comment(None)
        self.assertRaises(KeyError, self._config.get_comment)

    def test_displayname(self):
        self.assertRaises(KeyError, self._config.get_displayname)
        self._config.set_displayname('DiSpLaYName')
        self.assertEqual('DiSpLaYName', self._config.get_displayname())
        self._config.set_displayname(None)
        self.assertRaises(KeyError, self._config.get_displayname)

    def test_description(self):
        self.assertRaises(KeyError, self._config.get_description)
        self._config.set_description('this is a description')
        self.assertEqual(
            'this is a description', self._config.get_description())
        self._config.set_description(None)
        self.assertRaises(KeyError, self._config.get_description)

    def test_order(self):
        self.assertRaises(KeyError, self._config.get_order)
        self._config.set_order('this is a order')
        self.assertEqual('this is a order', self._config.get_order())
        self._config.set_order(None)
        self.assertRaises(KeyError, self._config.get_order)


class FileMetadataTests(TestCase, MetadataTests):

    def setUp(self):
        super(FileMetadataTests, self).setUp()
        self._config = FileBasedCollectionMetadata()


class RepoMetadataTests(TestCase, MetadataTests):

    def setUp(self):
        super(RepoMetadataTests, self).setUp()
        self._repo = dulwich.repo.MemoryRepo()
        self._config = RepoCollectionMetadata(self._repo)



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Collations."""

from typing import Callable


class UnknownCollation(Exception):

    def __init__(self, collation: str):
        super(UnknownCollation, self).__init__(
            "Collation %r is not supported" % collation)
        self.collation = collation


collations = {
    'i;ascii-casemap': lambda a, b: (a.decode('ascii').upper() ==
                                     b.decode('ascii').upper()),
    'i;octet': lambda a, b: a == b,
}


def get_collation(name: str) -> Callable[[str, str], bool]:
    """Get a collation by name.

    :param name: Collation name
    :raises UnknownCollation: If the collation is not supported
    """
    try:
        return collations[name]
    except KeyError:
        raise UnknownCollation(name)



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""WSGI wrapper for xandikos.
"""

import logging
import os
import posixpath

from xandikos.web import (
    XandikosBackend,
    XandikosApp,
    WELLKNOWN_DAV_PATHS,
)


class WellknownRedirector(object):
    """Redirect paths under .well-known/ to the appropriate paths."""

    def __init__(self, inner_app, dav_root):
        self._inner_app = inner_app
        self._dav_root = dav_root

    def __call__(self, environ, start_response):
        # See https://tools.ietf.org/html/rfc6764
        path = posixpath.normpath(
            environ['SCRIPT_NAME'] + environ['PATH_INFO'])
        if path in WELLKNOWN_DAV_PATHS:
            start_response('302 Found', [
                ('Location', self._dav_root)])
            return []
        return self._inner_app(environ, start_response)


if __name__ == '__main__':
    backend = XandikosBackend(path=os.environ['XANDIKOSPATH'])
    if not os.path.isdir(backend.path):
        if os.getenv('AUTOCREATE'):
            os.makedirs(os.environ['XANDIKOSPATH'])
        else:
            logging.warning('%r does not exist.', backend.path)

    current_user_principal = os.environ.get('CURRENT_USER_PRINCIPAL', '/user/')
    if not backend.get_resource(current_user_principal):
        if os.getenv('AUTOCREATE'):
            backend.create_principal(
                current_user_principal,
                create_defaults=os.environ['AUTOCREATE'] == 'defaults')
        else:
            logging.warning(
                'default user principal \'%s\' does not exist. '
                'Create directory %s or set AUTOCREATE variable?',
                current_user_principal, backend._map_to_file_path(
                    current_user_principal))

    backend._mark_as_principal(current_user_principal)
    app = XandikosApp(backend, current_user_principal)



# Xandikos
# Copyright (C) 2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""ICalendar file handling.

"""

import datetime
import logging

from icalendar.cal import Calendar, component_factory
from icalendar.prop import (
    vDatetime,
    vDDDTypes,
    vText,
)
from xandikos.store import (
    Filter,
    File,
    InvalidFileContents,
)

from . import (
    collation as _mod_collation,
)

# TODO(jelmer): Populate this further based on
# https://tools.ietf.org/html/rfc5545#3.3.11
_INVALID_CONTROL_CHARACTERS = ['\x0c', '\x01']


class MissingProperty(Exception):

    def __init__(self, property_name):
        super(MissingProperty, self).__init__(
            "Property %r missing" % property_name)
        self.property_name = property_name


def validate_calendar(cal, strict=False):
    """Validate a calendar object.

    :param cal: Calendar object
    :return: iterator over error messages
    """
    for error in validate_component(cal, strict=strict):
        yield error


def create_subindexes(indexes, base):
    ret = {}
    for k, v in indexes.items():
        if k is not None and k.startswith(base + '/'):
            ret[k[len(base) + 1:]] = v
        elif k == base:
            ret[None] = v
    return ret


def validate_component(comp, strict=False):
    """Validate a calendar component.

    :param comp: Calendar component
    """
    # Check text fields for invalid characters
    for (name, value) in comp.items():
        if isinstance(value, vText):
            for c in _INVALID_CONTROL_CHARACTERS:
                if c in value:
                    yield "Invalid character %s in field %s" % (
                        c.encode('unicode_escape'), name)
    if strict:
        for required in comp.required:
            try:
                comp[required]
            except KeyError:
                yield "Missing required field %s" % required
    for subcomp in comp.subcomponents:
        for error in validate_component(subcomp, strict=strict):
            yield error


def calendar_component_delta(old_cal, new_cal):
    """Find the differences between components in two calendars.

    :param old_cal: Old calendar (can be None)
    :param new_cal: New calendar (can be None)
    :yield: (old_component, new_component) tuples (either can be None)
    """
    by_uid = {}
    by_content = {}
    by_idx = {}
    idx = 0
    for component in getattr(old_cal, "subcomponents", []):
        try:
            by_uid[component["UID"]] = component
        except KeyError:
            by_content[component.to_ical()] = True
            by_idx[idx] = component
            idx += 1
    idx = 0
    for component in new_cal.subcomponents:
        try:
            old_component = by_uid.pop(component["UID"])
        except KeyError:
            if not by_content.pop(component.to_ical(), None):
                # Not previously present
                yield (by_idx.get(idx, component_factory[component.name]()),
                       component)
            by_idx.pop(idx, None)
        else:
            yield (old_component, component)
    for old_component in by_idx.values():
        yield (old_component, component_factory[old_component.name]())


def calendar_prop_delta(old_component, new_component):
    fields = set([field for field in old_component or []] +
                 [field for field in new_component or []])
    for field in fields:
        old_value = old_component.get(field)
        new_value = new_component.get(field)
        if (
            getattr(old_value, 'to_ical', None) is None or
            getattr(new_value, 'to_ical', None) is None or
            old_value.to_ical() != new_value.to_ical()
        ):
            yield (field, old_value, new_value)


def describe_component(component):
    if component.name == "VTODO":
        try:
            return "task '%s'" % component["SUMMARY"]
        except KeyError:
            return "task"
    else:
        try:
            return component["SUMMARY"]
        except KeyError:
            return "calendar item"


DELTA_IGNORE_FIELDS = set(["LAST-MODIFIED", "SEQUENCE", "DTSTAMP", "PRODID",
                           "CREATED", "COMPLETED", "X-MOZ-GENERATION",
                           "X-LIC-ERROR", "UID"])


def describe_calendar_delta(old_cal, new_cal):
    """Describe the differences between two calendars.

    :param old_cal: Old calendar (can be None)
    :param new_cal: New calendar (can be None)
    :yield: Lines describing changes
    """
    # TODO(jelmer): Extend
    for old_component, new_component in calendar_component_delta(old_cal,
                                                                 new_cal):
        if not new_component:
            yield "Deleted %s" % describe_component(old_component)
            continue
        description = describe_component(new_component)
        if not old_component:
            yield "Added %s" % describe_component(new_component)
            continue
        for field, old_value, new_value in calendar_prop_delta(old_component,
                                                               new_component):
            if field.upper() in DELTA_IGNORE_FIELDS:
                continue
            if (
                old_component.name.upper() == "VTODO" and
                field.upper() == "STATUS"
            ):
                if new_value is None:
                    yield "status of %s deleted" % description
                else:
                    human_readable = {
                        "NEEDS-ACTION": "needing action",
                        "COMPLETED": "complete",
                        "CANCELLED": "cancelled"}
                    yield "%s marked as %s" % (
                        description,
                        human_readable.get(new_value.upper(), new_value))
            elif field.upper() == 'DESCRIPTION':
                yield "changed description of %s" % description
            elif field.upper() == 'SUMMARY':
                yield "changed summary of %s" % description
            elif field.upper() == 'LOCATION':
                yield "changed location of %s to %s" % (description, new_value)
            elif (old_component.name.upper() == "VTODO" and
                  field.upper() == "PERCENT-COMPLETE" and
                  new_value is not None):
                yield "%s marked as %d%% completed." % (
                    description, new_value)
            elif field.upper() == 'DUE':
                yield "changed due date for %s from %s to %s" % (
                    description, old_value.dt if old_value else 'none',
                    new_value.dt if new_value else 'none')
            elif field.upper() == 'DTSTART':
                yield "changed start date/time of %s from %s to %s" % (
                    description, old_value.dt if old_value else 'none',
                    new_value.dt if new_value else 'none')
            elif field.upper() == 'DTEND':
                yield "changed end date/time of %s from %s to %s" % (
                    description, old_value.dt if old_value else 'none',
                    new_value.dt if new_value else 'none')
            elif field.upper() == 'CLASS':
                yield "changed class of %s from %s to %s" % (
                    description, old_value.lower() if old_value else 'none',
                    new_value.lower() if new_value else 'none')
            else:
                yield "modified field %s in %s" % (field, description)
                logging.debug("Changed %s/%s or %s/%s from %s to %s.",
                              old_component.name, field, new_component.name,
                              field, old_value, new_value)


def apply_time_range_vevent(start, end, comp, tzify):
    dtstart = comp.get('DTSTART')
    if not dtstart:
        raise MissingProperty('DTSTART')

    if not (end > tzify(dtstart.dt)):
        return False

    dtend = comp.get('DTEND')
    if dtend:
        if tzify(dtend.dt) < tzify(dtstart.dt):
            logging.debug('Invalid DTEND < DTSTART')
        return (start < tzify(dtend.dt))

    duration = comp.get('DURATION')
    if duration:
        return (start < tzify(dtstart.dt) + duration.dt)
    if getattr(dtstart.dt, 'time', None) is not None:
        return (start <= tzify(dtstart.dt))
    else:
        return (start < (tzify(dtstart.dt) + datetime.timedelta(1)))


def apply_time_range_vjournal(start, end, comp, tzify):
    dtstart = comp.get('DTSTART')
    if not dtstart:
        raise MissingProperty('DTSTART')

    if not (end > tzify(dtstart.dt)):
        return False

    if getattr(dtstart.dt, 'time', None) is not None:
        return (start <= tzify(dtstart.dt))
    else:
        return (start < (tzify(dtstart.dt) + datetime.timedelta(1)))


def apply_time_range_vtodo(start, end, comp, tzify):
    dtstart = comp.get('DTSTART')
    due = comp.get('DUE')

    # See RFC4719, section 9.9
    if dtstart:
        duration = comp.get('DURATION')
        if duration and not due:
            return (
                start <= tzify(dtstart.dt) + duration.dt and
                (end > tzify(dtstart.dt) or
                 end >= tzify(dtstart.dt) + duration.dt)
            )
        elif due and not duration:
            return (
                (start <= tzify(dtstart.dt) or
                 start < tzify(due.dt)) and
                (end > tzify(dtstart.dt) or
                 end < tzify(due.dt))
            )
        else:
            return (start <= tzify(dtstart.dt) and
                    end > tzify(dtstart.dt))

    if due:
        return start < tzify(due.dt) and end >= tzify(due.dt)

    completed = comp.get('COMPLETED')
    created = comp.get('CREATED')
    if completed:
        if created:
            return (
                (start <= tzify(created.dt) or
                 start <= tzify(completed.dt)) and
                (end >= tzify(created.dt) or
                 end >= tzify(completed.dt))
            )
        else:
            return (
                start <= tzify(completed.dt) and
                end >= tzify(completed.dt)
            )
    elif created:
        return end >= tzify(created.dt)
    else:
        return True


def apply_time_range_vfreebusy(start, end, comp, tzify):
    dtstart = comp.get('DTSTART')
    dtend = comp.get('DTEND')
    if dtstart and dtend:
        return (
            start <= tzify(dtend.dt) and
            end > tzify(dtstart.dt)
        )

    for period in comp.get('FREEBUSY', []):
        if start < period.end and end > period.start:
            return True

    return False


def apply_time_range_valarm(start, end, comp, tzify):
    raise NotImplementedError(apply_time_range_valarm)


class PropertyTimeRangeMatcher(object):

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.start, self.end)

    def match(self, prop, tzify):
        dt = tzify(prop.dt)
        return (dt >= self.start and dt <= self.end)

    def match_indexes(self, prop, tzify):
        return any(self.match(vDDDTypes(vDatetime.from_ical(p)), tzify)
                   for p in prop[None])


class ComponentTimeRangeMatcher(object):

    all_props = [
        'DTSTART', 'DTEND', 'DURATION', 'CREATED', 'COMPLETED', 'DUE',
        'FREEBUSY']

    # According to https://tools.ietf.org/html/rfc4791, section 9.9 these
    # are the properties to check.
    component_handlers = {
        'VEVENT': apply_time_range_vevent,
        'VTODO': apply_time_range_vtodo,
        'VJOURNAL': apply_time_range_vjournal,
        'VFREEBUSY': apply_time_range_vfreebusy,
        'VALARM': apply_time_range_valarm}

    def __init__(self, start, end, comp=None):
        self.start = start
        self.end = end
        self.comp = comp

    def __repr__(self):
        if self.comp is not None:
            return "%s(%r, %r, comp=%r)" % (
                self.__class__.__name__, self.start, self.end, self.comp)
        else:
            return "%s(%r, %r)" % (
                self.__class__.__name__, self.start, self.end)

    def match(self, comp, tzify):
        try:
            component_handler = self.component_handlers[comp.name]
        except KeyError:
            logging.warning('unknown component %r in time-range filter',
                            comp.name)
            return False
        return component_handler(self.start, self.end, comp, tzify)

    def match_indexes(self, indexes, tzify):
        vs = {}
        for name, value in indexes.items():
            if name and name[2:] in self.all_props:
                if value and value[0]:
                    if not isinstance(value[0], vDDDTypes):
                        vs[name[2:]] = vDDDTypes(vDatetime.from_ical(value[0]))
                    else:
                        vs[name[2:]] = value[0]

        try:
            component_handler = self.component_handlers[self.comp]
        except KeyError:
            logging.warning('unknown component %r in time-range filter',
                            self.comp)
            return False
        return component_handler(self.start, self.end, vs, tzify)

    def index_keys(self):
        if self.comp == 'VEVENT':
            props = ['DTSTART', 'DTEND', 'DURATION']
        elif self.comp == 'VTODO':
            props = ['DTSTART', 'DUE', 'DURATION', 'CREATED', 'COMPLETED']
        elif self.comp == 'VJOURNAL':
            props = ['DTSTART']
        elif self.comp == 'VFREEBUSY':
            props = ['DTSTART', 'DTEND', 'FREEBUSY']
        elif self.comp == 'VALARM':
            raise NotImplementedError
        else:
            props = self.all_props
        return [['P=' + prop] for prop in props]


class TextMatcher(object):

    def __init__(self, text, collation=None, negate_condition=False):
        if isinstance(text, str):
            text = text.encode()
        self.text = text
        if collation is None:
            collation = 'i;ascii-casemap'
        self.collation = _mod_collation.get_collation(collation)
        self.negate_condition = negate_condition

    def __repr__(self):
        return '%s(%r, collation=%r, negate_condition=%r)' % (
            self.__class__.__name__, self.text, self.collation,
            self.negate_condition)

    def match_indexes(self, indexes):
        return any(self.match(k) for k in indexes[None])

    def match(self, prop):
        if isinstance(prop, vText):
            prop = prop.encode()
        matches = self.collation(self.text, prop)
        if self.negate_condition:
            return not matches
        else:
            return matches


class ComponentFilter(object):

    def __init__(self, name, children=None, is_not_defined=False,
                 time_range=None):
        self.name = name
        self.children = children
        self.is_not_defined = is_not_defined
        self.time_range = time_range
        self.children = children or []

    def __repr__(self):
        return '%s(%r, children=%r, is_not_defined=%r, time_range=%r)' % (
            self.__class__.__name__, self.name, self.children,
            self.is_not_defined, self.time_range)

    def filter_subcomponent(self, name, is_not_defined=False,
                            time_range=None):
        ret = ComponentFilter(
            name=name, is_not_defined=is_not_defined, time_range=time_range)
        self.children.append(ret)
        return ret

    def filter_property(self, name, is_not_defined=False, time_range=None):
        ret = PropertyFilter(
            name=name, is_not_defined=is_not_defined, time_range=time_range)
        self.children.append(ret)
        return ret

    def filter_time_range(self, start, end):
        self.time_range = ComponentTimeRangeMatcher(
            start, end, comp=self.name)
        return self.time_range

    def match(self, comp, tzify):
        # From https://tools.ietf.org/html/rfc4791, 9.7.1:
        # A CALDAV:comp-filter is said to match if:

        # 2. The CALDAV:comp-filter XML element contains a
        # CALDAV:is-not-defined XML element and the calendar object or calendar
        # component type specified by the "name" attribute does not exist in
        # the current scope;
        if self.is_not_defined:
            return comp.name != self.name

        # 1: The CALDAV:comp-filter XML element is empty and the calendar
        # object or calendar component type specified by the "name" attribute
        # exists in the current scope;
        if comp.name != self.name:
            return False

        # 3. The CALDAV:comp-filter XML element contains a CALDAV:time-range
        # XML element and at least one recurrence instance in the targeted
        # calendar component is scheduled to overlap the specified time range
        if (self.time_range is not None and
                not self.time_range.match(comp, tzify)):
            return False

        # ... and all specified CALDAV:prop-filter and CALDAV:comp-filter child
        # XML elements also match the targeted calendar component;
        for child in self.children:
            if isinstance(child, ComponentFilter):
                if not any(child.match(c, tzify) for c in comp.subcomponents):
                    return False
            elif isinstance(child, PropertyFilter):
                if not child.match(comp, tzify):
                    return False
            else:
                raise TypeError(child)

        return True

    def _implicitly_defined(self):
        return any(not getattr(child, 'is_not_defined', False)
                   for child in self.children)

    def match_indexes(self, indexes, tzify):
        myindex = 'C=' + self.name
        if self.is_not_defined:
            return not bool(indexes[myindex])

        subindexes = create_subindexes(indexes, myindex)

        if (self.time_range is not None and
                not self.time_range.match_indexes(subindexes, tzify)):
            return False

        for child in self.children:
            if not child.match_indexes(subindexes, tzify):
                return False

        if not self._implicitly_defined():
            return bool(indexes[myindex])

        return True

    def index_keys(self):
        mine = 'C=' + self.name
        for child in (
                self.children +
                ([self.time_range] if self.time_range else [])):
            for tl in child.index_keys():
                yield [(mine + '/' + child_index) for child_index in tl]
        if not self._implicitly_defined():
            yield [mine]


class PropertyFilter(object):

    def __init__(self, name, children=None, is_not_defined=False,
                 time_range=None):
        self.name = name
        self.is_not_defined = is_not_defined
        self.children = children or []
        self.time_range = time_range

    def __repr__(self):
        return '%s(%r, children=%r, is_not_defined=%r, time_range=%r)' % (
            self.__class__.__name__, self.name, self.children,
            self.is_not_defined, self.time_range)

    def filter_parameter(self, name, is_not_defined=False):
        ret = ParameterFilter(name=name, is_not_defined=is_not_defined)
        self.children.append(ret)
        return ret

    def filter_time_range(self, start, end):
        self.time_range = PropertyTimeRangeMatcher(start, end)
        return self.time_range

    def filter_text_match(self, text, collation=None, negate_condition=False):
        ret = TextMatcher(
            text, collation=collation, negate_condition=negate_condition)
        self.children.append(ret)
        return ret

    def match(self, comp, tzify):
        # From https://tools.ietf.org/html/rfc4791, 9.7.2:
        # A CALDAV:comp-filter is said to match if:

        # The CALDAV:prop-filter XML element contains a CALDAV:is-not-defined
        # XML element and no property of the type specified by the "name"
        # attribute exists in the enclosing calendar component;

        if self.is_not_defined:
            return self.name not in comp

        try:
            prop = comp[self.name]
        except KeyError:
            return False

        if self.time_range and not self.time_range.match(prop, tzify):
            return False

        for child in self.children:
            if not child.match(prop):
                return False

        return True

    def match_indexes(self, indexes, tzify):
        myindex = 'P=' + self.name
        if self.is_not_defined:
            return not bool(indexes[myindex])
        subindexes = create_subindexes(indexes, myindex)
        if not self.children and not self.time_range:
            return bool(indexes[myindex])

        if (self.time_range is not None and
                not self.time_range.match_indexes(subindexes, tzify)):
            return False

        for child in self.children:
            if not child.match_indexes(subindexes):
                return False

        return True

    def index_keys(self):
        mine = 'P=' + self.name
        for child in self.children:
            if not isinstance(child, ParameterFilter):
                continue
            for tl in child.index_keys():
                yield [(mine + '/' + child_index) for child_index in tl]
        yield [mine]


class ParameterFilter(object):

    def __init__(self, name, children=None, is_not_defined=False):
        self.name = name
        self.is_not_defined = is_not_defined
        self.children = children or []

    def filter_text_match(self, text, collation=None, negate_condition=False):
        ret = TextMatcher(
            text, collation=collation, negate_condition=negate_condition)
        self.children.append(ret)
        return ret

    def match(self, prop):
        if self.is_not_defined:
            return self.name not in prop.params

        try:
            value = prop.params[self.name].encode()
        except KeyError:
            return False

        for child in self.children:
            if not child.match(value):
                return False
        return True

    def index_keys(self):
        yield ['A=' + self.name]

    def match_indexes(self, indexes):
        myindex = 'A=' + self.name
        if self.is_not_defined:
            return not bool(indexes[myindex])

        try:
            value = indexes[myindex][0]
        except IndexError:
            return False

        for child in self.children:
            if not child.match(value):
                return False
        return True


class CalendarFilter(Filter):
    """A filter that works on ICalendar files."""

    content_type = 'text/calendar'

    def __init__(self, default_timezone):
        self.tzify = lambda dt: as_tz_aware_ts(dt, default_timezone)
        self.children = []

    def filter_subcomponent(self, name, is_not_defined=False,
                            time_range=None):
        ret = ComponentFilter(
            name=name, is_not_defined=is_not_defined, time_range=time_range)
        self.children.append(ret)
        return ret

    def check(self, name, file):
        c = file.calendar
        if c is None:
            return False

        for child_filter in self.children:
            try:
                if not child_filter.match(file.calendar, self.tzify):
                    return False
            except MissingProperty as e:
                logging.warning(
                    'calendar_query: Ignoring calendar object %s, due '
                    'to missing property %s', name, e.property_name)
                return False
        return True

    def check_from_indexes(self, name, indexes):
        for child_filter in self.children:
            try:
                if not child_filter.match_indexes(
                        indexes, self.tzify):
                    return False
            except MissingProperty as e:
                logging.warning(
                    'calendar_query: Ignoring calendar object %s, due '
                    'to missing property %s', name, e.property_name)
                return False
        return True

    def index_keys(self):
        subindexes = []
        for child in self.children:
            subindexes.extend(child.index_keys())
        return subindexes

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.children)


class ICalendarFile(File):
    """Handle for ICalendar files."""

    content_type = 'text/calendar'

    def __init__(self, content, content_type):
        super(ICalendarFile, self).__init__(content, content_type)
        self._calendar = None

    def validate(self):
        """Verify that file contents are valid."""
        cal = self.calendar
        # TODO(jelmer): return the list of errors to the caller
        if cal.is_broken:
            raise InvalidFileContents(
                self.content_type, self.content,
                "Broken calendar file")
        errors = list(validate_calendar(cal, strict=False))
        if errors:
            raise InvalidFileContents(
                self.content_type, self.content,
                ", ".join(errors))

    def normalized(self):
        """Return a normalized version of the file."""
        return [self.calendar.to_ical()]

    @property
    def calendar(self):
        if self._calendar is None:
            try:
                self._calendar = Calendar.from_ical(b''.join(self.content))
            except ValueError as e:
                raise InvalidFileContents(
                    self.content_type, self.content, str(e))
        return self._calendar

    def describe_delta(self, name, previous):
        try:
            lines = list(describe_calendar_delta(
                previous.calendar if previous else None, self.calendar))
        except NotImplementedError:
            lines = []
        if not lines:
            lines = super(ICalendarFile, self).describe_delta(name, previous)
        return lines

    def describe(self, name):
        try:
            subcomponents = self.calendar.subcomponents
        except InvalidFileContents:
            pass
        else:
            for component in subcomponents:
                try:
                    return component["SUMMARY"]
                except KeyError:
                    pass
        return super(ICalendarFile, self).describe(name)

    def get_uid(self):
        """Extract the UID from a VCalendar file.

        :param cal: Calendar, possibly serialized.
        :return: UID
        """
        for component in self.calendar.subcomponents:
            try:
                return component["UID"]
            except KeyError:
                pass
        raise KeyError

    def _get_index(self, key):
        todo = [(self.calendar, key.split('/'))]
        rest = []
        while todo:
            (c, segments) = todo.pop(0)
            if segments and segments[0].startswith('C='):
                if c.name == segments[0][2:]:
                    if len(segments) > 1 and segments[1].startswith('C='):
                        todo.extend(
                            (comp, segments[1:]) for comp in c.subcomponents)
                    else:
                        rest.append((c, segments[1:]))

        for c, segments in rest:
            if not segments:
                yield True
            elif segments[0].startswith('P='):
                assert len(segments) == 1
                try:
                    yield c[segments[0][2:]]
                except KeyError:
                    pass
            else:
                raise AssertionError('segments: %r' % segments)


def as_tz_aware_ts(dt, default_timezone):
    if not getattr(dt, 'time', None):
        dt = datetime.datetime.combine(dt, datetime.time())
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=default_timezone)
    assert dt.tzinfo
    return dt
